#!/usr/bin/perl
#
# Manipulate /etc/inetd.conf
#
# Adding an entry:
#
#   inetconf.pl --add \
#		finger stream tcp nowait root /usr/sbin/tcpd in.fingerd
#   inetconf.pl --add \
#		rwall/1 dgram rpc/udp wait nobody.nogroup /usr/sbin/rpc.rwalld
#
# If the service already exists, the entry is overwritten.
#
# Notes:
#  *		The requested service name must be listed in /etc/services
#		or /etc/rpc (in the case of RPC servies).
#  *		The requested user and group must be listed in /etc/passwd
#		or /etc/group.
#  *		The script performs some consistency checks on the other
#		supplied arguments.
#
# Removing an entry:
#
#   inetconf.pl --remove finger
#   inetconf.pl --remove rwall/1
#   inetconf.pl --remove rwall
#
# Disabling an entry:
#
#   inetconf.pl --disable finger
#
#		Disables the entry by placing a # in front of it.
#		May also be used in conjunction with --install.
#
# Options:
#   --text	Puts the descriptive text in a comment in front of the
#		entry.
#   --config	Use a different inet.conf file (for debugging)
#

$conffile = "/etc/inetd.conf";

$addrem = 0;
while ($_ = shift) {
	last unless (/^--/);
	if ($_ eq "--add") {
		$addrem = 1;
		$add = 1;
	} elsif ($_ eq "--remove") {
		$addrem = 1;
		$add = 0;
	} elsif ($_ eq "--disable") {
		$disable = 1;
		unless($addrem) {
			$add = 0;
			$addrem = 1;
		}
	} elsif ($_ eq "--text") {
		$desc = "# " . shift;
	} elsif ($_ eq "--config") {
		$conffile = shift;
	} else {
		die "Unknown option $_\n";
	}
}

if (!$addrem) {
	die "Argument --add or --remove required\n";
}

$service = $_;

if ($service =~ /^$/) {
	die "Missing service argument\n";
}

if ($add) {
	&add($service, shift, shift, shift, shift, @ARGV);
} else {
	&remove($service);
}

0;

##################################################################
# Remove a service
##################################################################
sub remove {

	local($service) = @_;
	local(@svcnames);

	open(OLD, "<$conffile") || &fatal("Can't open $conffile: $!\n");
	open(NEW, ">$conffile.new") || &fatal("Can't open $conffile.new: $!\n");

	# FIXME:
	# We can't get the service aliases here because we don't know
	# whether it's RPC or ARPA.
	@svcnames = ($service);

	$leader = '';
	while (<OLD>) {
		if (&match_service($_, @svcnames)) {
			if ($disable) {
				$_ = "#" . $_;
			} else {
				$leader = '';
				next;
			}
		} elsif (/^#/) {
			$leader .= $_;
			next;
		}
		print NEW $leader;
		print NEW $_;
		$leader = '';
	}

	print NEW $leader;
	close OLD;
	close NEW;

	&install_new("$conffile.new", "$conffile",
		     "$service ($prot)", "removed");
}

##################################################################
# Add a service
##################################################################
sub add {

	local($service, $type, $prot, $wait, $user, @cmd) = @_;
	local($_, $leader, $added, $action);
	local(@svcnames);

	if ($type eq "stream") {
		if ($prot ne "tcp" && $prot ne "rpc/tcp" && $prot ne "unix") {
			&fatal("Bad protocol $prot for stream socket");
		}
	} elsif ($type eq "dgram") {
		if ($prot ne "udp" && $prot ne "rpc/udp" && $prot ne "unix") {
			&fatal("Bad protocol $prot for dgram socket");
		}
	} else {
		&fatal("Invalid socket type $type\n");
	}
	if ($wait ne "wait" && !$wait =~ /^nowait\.\d+$/ && $wait ne "nowait") {
		&fatal("Invalid wait option $wait");
	}
	if ($prot =~ /rpc\//) {
		local($rpcsvc) = $service;
		if (!($rpcsvc =~ s/\/[0-9]*$//)) {
			&fatal("No RPC version in service name $service");
		}
		&verify_service($rpcsvc, "/etc/rpc");
		@svcnames = ($service);
	} elsif ($prot eq "tcp" || $prot eq "udp") {
		@svcnames = &verify_service($service, "/etc/services");
	}

	# user.group or plain user
	if ($user =~ /(.*)\.(.*)/) {
		&verify_user($1, "/etc/passwd");
		&verify_user($2, "/etc/group");
	} else {
		&verify_user($user, "/etc/passwd");
	}

	open(OLD, "<$conffile") || &fatal("Can't open $conffile: $!\n");
	open(NEW, ">$conffile.new") || &fatal("Can't open $conffile.new: $!\n");

	$leader = '';
	$action = "added";
	$added  = 0;
	while (<OLD>) {
		study;
		if (&match_service($_, @svcnames)) {
			if ($desc eq '') {
				chop($leader);
				$desc = $leader;
			}
			$action = "modified";
			$leader = '';
			last;
		}
		if (/^#/) {
			$leader .= $_;
			next;
		}
		print NEW $leader;
		print NEW $_;
		$leader = '';
	}

	# Print comment leader and service line
	print NEW "$desc\n" unless ($desc eq '');
	printf(NEW "%s%-7s %-7s %-7s %-7s %-7s",
		$disable? "#" : "",
		$service, $type, $prot, $wait, $user);
	foreach (@cmd) {
		print NEW " $_";
	}
	print NEW "\n";

	while (<OLD>) {
		study;
		# Weed out duplicates
		if (&match_service($_, @svcnames)) {
			$leader = '';
			next;
		}
		if (/^#/) {
			$leader .= $_;
			next;
		}
		print NEW $leader;
		print NEW $_;
		$leader = '';
	}
	print NEW $leader;
	close OLD;
	close NEW;

	&install_new("$conffile.new", "$conffile",
		     "$service ($prot)", $action);
}

##################################################################
# Match an inetd.conf line against the given services
##################################################################
sub match_service {

	local($line, @svcnames) = @_;
	local($_);

	foreach (@svcnames) {
		return 1 if ($line =~ /^#*\s*$_\b/);
	}

	return 0;
}

##################################################################
# Check whether the named service is in /etc/services or /etc/rpc
##################################################################
sub verify_service {

	local($svc, $file) = @_;
	local($_, @list);

	open(SVC, $file) || &fatal("Can't open $file: $!\n");
	while (<SVC>) {
		s/#.*//;
		@list = split;
		splice(@list, 1, 1);
		foreach (@list) {
			if ($svc eq $_) {
				close (SVC);
				return @list;
			}
		}
	}

	&fatal("Service \"$svc\" not found in $file");
}

##################################################################
# Make sure the requested user is actually in the system
# Note that we don't deal with NIS and other monstrosities,
# since system accounts should always be local.
##################################################################
sub verify_user {

	local($name, $file) = @_;
	local($_);

	open(PASS, $file) || &fatal("Can't open $file: $!");
	while (<PASS>) {
		if (/^$name:/i) {
			close (PASS);
			return 1;
		}
	}

	&fatal("Couldn't find $name in $file");
}

##################################################################
# Print a fatal message
##################################################################
sub fatal {

	local($msg) = @_;

	print STDERR "Can't create inetd.conf entry for $service:\n";
	print STDERR "$msg\n";

	exit 1;
}

##################################################################
# Replace inetd.conf with the changed version
##################################################################
sub install_new {

	local($newfile, $file, $service, $action) = @_;

	# Do not check error code... leave the .orig file as-is
	# if it exists.
	link($file, "$file.orig")
		unless ($checked_out);

	rename($newfile, $file)
		|| &fatal("Can't replace old $file: $!\n");
	system("killall -HUP inetd 2>&1 >/dev/null");
	&checkin($file, $service, $action);
}

##################################################################
# Check out the original inetd.conf file
##################################################################
sub checkout {

	local($file) = @_;
	local($dir, $base);

	if ($file =~ /(.*)\/(.*)/) {
		$dir  = $1;
		$base = $2;
	} else {
		$dir  = ".";
		$base = $file;
	}

	if (-d "$dir/RCS/$base,v" && ! -w $file && -c "/usr/bin/co") {
		system("co -l $file")
			|| &fatal("Can't check out $file from RCS\n");
		$checked_out = 1;
	}
}

sub checkin {

	local($file, $service, $action) = @_;
	local($dir, $base);

	unless ($checked_out) {
		# Do not check error code... leave the .orig file as-is
		# if it exists.
		link($file, "$file.orig");
		return;
	}

	if ($file =~ /(.*)\/(.*)/) {
		$dir  = $1;
		$base = $2;
	} else {
		$dir  = ".";
		$base = $file;
	}

	system("ci -u -m\"LISA: $action $service")
		|| &fatal("Can't check in $file in RCS\n");
}
