/*
 * getty.c -- getty main file
 * Copyright (C) 1998,1999 Torsten Duwe <duwe@caldera.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <strings.h>		/* for ffs() */
#include <syslog.h>
#include <errno.h>

struct termios ti;
long speed = 9600, speedcode = 0;
unsigned parity = 0;

extern int cgetent(char **buf, char **db_array, const char *name);
extern int cgetstr(char *buf, const char *cap, char **str);
extern int cgetmatch(const char *buf, const char *name);
extern int cgetnum(char *buf, const char *cap, long *num);

void stty_sane(int full);

int issuefd;
struct utsname hname;
#define MAXLOGNAME 80
char logname[MAXLOGNAME+2];

char * promptstr = "\r\n%h login: ",
  * device = "/dev/tty",
  * device_fullpath = 0,
  * termtype = 0,
  * login_prog = "/bin/login",
  * issuepath = "/etc/issue",
#ifdef CONFIGURABLE
  * conffiles[] = { "/etc/gettytab", 0 },
  * linespeed = "default",
  * ttycap = 0,
#endif
  * progname,
  * hostname = hname.nodename;

int 
main (int argc, char ** argv, char ** envp)
{
  char ** argp;
  int argsleft;
  int r, fd, ldisc = 0;
  int i, c;
  char * p;
  
  progname = argv[0];
  openlog("getty", LOG_PID, LOG_AUTHPRIV);

  while ((r = getopt(argc, argv, "a:c:d:hr:t:w:")) != EOF)
    switch ((char)r)
      {
      case 'a':
	break;
      case 'c':
#ifdef notyet
	checkonly = 1;
	conffiles[0] = optarg;
#endif
	break;
      case 'd':
	break;
      case 'h':
#ifdef notyet
	hup = 0;
#endif
	break;
      case 'r':
	break;
      case 't':
	break;
      case 'w':
	break;
	
      default:
	break;
      }
  argp = argv + optind;
  argsleft = argc - optind;

  if (argsleft < 1)
    {
      syslog(LOG_ERR, "must specify device to work on.");
      exit(1);
    }

  argsleft--;
  device = device_fullpath = *argp++;

  if (argsleft)
    {
#ifdef CONFIGURABLE
      linespeed = *argp;
#endif
      argsleft--, argp++;

      if (argsleft)
	{
	  argsleft--;
	  termtype = *argp++;
	}
    }

  if (argsleft)
    syslog(LOG_INFO, "last %d args ignored", argsleft);

  c = getdtablesize();
  for (i=c-1; i>0; i--)
    (void)close(i);

  (void)setsid();
  
  if (strcmp("-", device))
    {
      close(0);
      
      if (strncmp("/dev/", device, 5))
	{
	  device_fullpath = malloc(6+strlen(device));
	  if (!device_fullpath)
	    {
#if 0	/* no stderr */
	      err(1, "cannot malloc device pathname");
#endif
	      exit(1);
	    }
	  strcpy(device_fullpath, "/dev/");
	  strcat(device_fullpath, device);
	}

      if ((fd = open(device_fullpath, O_RDWR|O_NONBLOCK, 0)) < 0)
	{
#if 0	/* no stderr */
	  err(1, "open device %s", device_fullpath);
#endif
	  exit(1);
	}

      if(device_fullpath != device)
	free(device_fullpath);

      if (fd > 0)
	{
	  dup2(fd,0);
	  close(fd);
	}
    }
  dup2(0,1);
  dup2(0,2);

  uname(&hname);

#ifdef CONFIGURABLE
  if (cgetent(&ttycap, conffiles, linespeed) < 0)
    ttycap = 0;
  else
    {
      char * tmp;
      
      if (cgetstr(ttycap, "hn", &tmp) > 0)
	hostname = strdup(tmp);

      if (cgetstr(ttycap, "im", &tmp) > 0)
	issuepath = strdup(tmp);

      if (cgetstr(ttycap, "lm", &tmp) > 0)
	promptstr = strdup(tmp);

      if (cgetstr(ttycap, "tt", &tmp) > 0 && !termtype)	/* cmdline overrides */
	termtype = strdup(tmp);

      if (cgetmatch(ttycap, "ep") == 0)
	parity = PARENB;
      else if (cgetmatch(ttycap, "np") == 0)
	parity = 0;
      else
	parity = 0;		/* default */

      if (cgetnum(ttycap, "sp", &speed) < 0)
	speed = 9600;
    }
#endif

  if (!termtype)
    termtype = "vt100";

#ifdef __linux__
  /* Linux CBAUD generation -- supports 300Bd x powers of 2, up to 38400 */
  if (speed % 300)
    speed = 9600;
  speedcode = speed / 300;
  if (speedcode & (speedcode -1))	/* more than 1 bit set */
    speed = 9600, speedcode = 32;	/* 9600 / 300 */
  speedcode = ffs(speedcode) + B200;	/* B200 + 1 from ffs(1) = B300 */
  if (speedcode >= 10) speedcode++;	/* skip B1800 */
  if (speedcode > 15)			/* upper limit for this calculation */
    speed = 38400, speedcode = 15;
#endif

#if defined (__NetBSD__)
  ldisc = TTYDISC;
#elif defined (__linux__)
  ldisc = N_TTY;
#endif

  ioctl(0, TIOCSETD, &ldisc);
  
  if (tcgetattr(0, &ti) < 0)
    syslog(LOG_ERR, "cannot get terminal state");

  cfmakeraw(&ti);

  if (tcsetattr(0, TCSANOW, &ti) < 0)
    syslog(LOG_ERR, "cannot set terminal state");

#ifdef TIOCSCTTY
  ioctl(0, TIOCSCTTY, 1);
#endif

#ifdef __linux__
  /*  vhangup(); -- login does this anyway. */
#endif
  
  stty_sane(0);

  if (tcsetattr(0, TCSANOW, &ti) < 0)
    syslog(LOG_ERR, "cannot set terminal state");

  fcntl(0, F_SETFL, O_RDWR);

  if ((issuefd = open(issuepath, O_RDONLY, 0)) >= 0)
    {
      char buf[256];
      int n;
      /* There's no buffered IO going on yet, so unbuffered is ok. */
      while ((n = read(issuefd, buf, 256)) > 0)
	write(1, buf, n);
      close(issuefd);
    }

  do
    {
#ifdef CONFIGURABLE
      time_t now;
      char timestr[20];
      char speedstr[8];
      char * ps;
      char * promptargs[4];
      int pargc = 0;

      (void)time(&now);
      strftime(timestr, 20, "%Y-%m-%d %T", localtime(&now));
      snprintf(speedstr, 8, "%ld", speed);

      ps = strdup(promptstr);
      for (p = ps; *p && pargc <= 4;)
	if (*p++ == '%')
	  switch (*p)
	    {
	    common:
	      *p = 's';
	      pargc++;
	      break;

	    case 'h':
	      promptargs[pargc] = hostname;
	      goto common;
	    case 't':
	      promptargs[pargc] = device;
	      goto common;
	    case 'd':
	      promptargs[pargc] = timestr;
	      goto common;
	    case 'b':
	      promptargs[pargc] = speedstr;
	      goto common;
	    default:
	      break;
	    }
      
      printf(ps, promptargs[0], promptargs[1], promptargs[2], promptargs[3]);
      free(ps);
#else
      printf(promptstr, hname.nodename);
#endif

      p = logname;
      if (fgets(logname, MAXLOGNAME, stdin) == 0) 
	continue;

      /* parse the logname -- currently only the trailing \n is deleted. */
      do {
	if (*p == '\n') 
	  *p = 0;
	if (*p)
	  p++;
      } while (*p && p < logname + MAXLOGNAME);
    }
  while (p <= logname);		/* only the terminating \0 => empty */

  setenv("TERM", termtype, 1);

  stty_sane(1);
  if (tcsetattr(0, TCSANOW, &ti) < 0)
    syslog(LOG_ERR, "cannot set terminal state(2)");

  closelog();

  execl(login_prog, login_prog, logname, 0);

				/* NOTREACHED */
  return(0);
}

void
stty_sane(int full)
{
#ifdef __linux__
#define OXTABS XTABS
#endif

  ti.c_iflag = IGNBRK|PARMRK|INPCK|ICRNL;
  ti.c_oflag = OPOST|ONLCR|OXTABS;
  ti.c_cflag = (parity?CS7:CS8)|CSTOPB|CREAD|HUPCL|CLOCAL|CRTSCTS \
    |parity|speedcode;		/* speedcode f. Linux only, otherwise 0 */
  ti.c_lflag = ICANON|ECHO|ECHOE|ECHOK; /* |TOSTOP; */

#if !defined(__linux__) || !defined(__sparc__)
  ti.c_ispeed = ti.c_ospeed = speed;
#endif

  if (full)
    ti.c_lflag |= ISIG;

	/* XXX This assumes ASCII ! */

  ti.c_cc[VEOF]		= 'D' & 0x1f;
  ti.c_cc[VERASE]	=       0x7f;
  ti.c_cc[VKILL]	= 'H' & 0x1f;
  ti.c_cc[VWERASE]	= 'W' & 0x1f;
  ti.c_cc[VREPRINT]	= 'R' & 0x1f;

  ti.c_cc[VINTR]	= 'C' & 0x1f;
  ti.c_cc[VQUIT]	= '\\'& 0x1f;
  ti.c_cc[VSUSP]	= 'Z' & 0x1f;
  ti.c_cc[VSTART]	= 'Q' & 0x1f;
  ti.c_cc[VSTOP]	= 'S' & 0x1f;
  ti.c_cc[VLNEXT]	= 'V' & 0x1f;

  ti.c_cc[VMIN]		= 1;
  ti.c_cc[VTIME]	= 0;
}
